# coding: utf-8
"""
Exploration on COVID data.
"""

__version__ = "0.1.0"
__author__ = "Xavier Dupré, ..."


def check(verbose=1):
    """
    Runs a couple of functions to check the module is working.

    :param verbose: 0 to hide the standout output
    :return: list of dictionaries, result of each test
    """
    return []
